library angular2.src.directives.ng_class;

import "package:angular2/src/facade/lang.dart"
    show isPresent, isString, StringWrapper, isBlank;
import "package:angular2/metadata.dart" show Directive, LifecycleEvent;
import "package:angular2/core.dart" show ElementRef;
import "package:angular2/src/render/api.dart" show Renderer;
import "package:angular2/change_detection.dart"
    show KeyValueDiffer, IterableDiffer, IterableDiffers, KeyValueDiffers;
import "package:angular2/src/facade/collection.dart"
    show ListWrapper, StringMapWrapper, isListLikeIterable;

/**
 * Adds and removes CSS classes based on an {expression} value.
 *
 * The result of expression is used to add and remove CSS classes using the following logic,
 * based on expression's value type:
 * - {string} - all the CSS classes (space - separated) are added
 * - {Array} - all the CSS classes (Array elements) are added
 * - {Object} - each key corresponds to a CSS class name while values
 * are interpreted as {boolean} expression. If a given expression
 * evaluates to {true} a corresponding CSS class is added - otherwise
 * it is removed.
 *
 * # Example:
 *
 * ```
 * <div class="message" [ng-class]="{error: errorCount > 0}">
 *     Please check errors.
 * </div>
 * ```
 */
@Directive(
    selector: "[ng-class]",
    lifecycle: const [LifecycleEvent.onCheck, LifecycleEvent.onDestroy],
    properties: const ["rawClass: ng-class", "initialClasses: class"])
class NgClass {
  IterableDiffers _iterableDiffers;
  KeyValueDiffers _keyValueDiffers;
  ElementRef _ngEl;
  Renderer _renderer;
  dynamic _differ;
  String _mode;
  var _initialClasses = [];
  var _rawClass;
  NgClass(this._iterableDiffers, this._keyValueDiffers, this._ngEl,
      this._renderer) {}
  set initialClasses(v) {
    this._applyInitialClasses(true);
    this._initialClasses = isPresent(v) && isString(v) ? v.split(" ") : [];
    this._applyInitialClasses(false);
    this._applyClasses(this._rawClass, false);
  }

  set rawClass(v) {
    this._cleanupClasses(this._rawClass);
    if (isString(v)) {
      v = v.split(" ");
    }
    this._rawClass = v;
    if (isPresent(v)) {
      if (isListLikeIterable(v)) {
        this._differ = this._iterableDiffers.find(v).create(null);
        this._mode = "iterable";
      } else {
        this._differ = this._keyValueDiffers.find(v).create(null);
        this._mode = "keyValue";
      }
    } else {
      this._differ = null;
    }
  }

  void onCheck() {
    if (isPresent(this._differ)) {
      var changes = this._differ.diff(this._rawClass);
      if (isPresent(changes)) {
        if (this._mode == "iterable") {
          this._applyIterableChanges(changes);
        } else {
          this._applyKeyValueChanges(changes);
        }
      }
    }
  }

  void onDestroy() {
    this._cleanupClasses(this._rawClass);
  }

  void _cleanupClasses(rawClassVal) {
    this._applyClasses(rawClassVal, true);
    this._applyInitialClasses(false);
  }

  void _applyKeyValueChanges(dynamic changes) {
    changes.forEachAddedItem((record) {
      this._toggleClass(record.key, record.currentValue);
    });
    changes.forEachChangedItem((record) {
      this._toggleClass(record.key, record.currentValue);
    });
    changes.forEachRemovedItem((record) {
      if (record.previousValue) {
        this._toggleClass(record.key, false);
      }
    });
  }

  void _applyIterableChanges(dynamic changes) {
    changes.forEachAddedItem((record) {
      this._toggleClass(record.item, true);
    });
    changes.forEachRemovedItem((record) {
      this._toggleClass(record.item, false);
    });
  }

  _applyInitialClasses(bool isCleanup) {
    ListWrapper.forEach(this._initialClasses, (className) {
      this._toggleClass(className, !isCleanup);
    });
  }

  _applyClasses(rawClassVal, bool isCleanup) {
    if (isPresent(rawClassVal)) {
      if (isListLikeIterable(rawClassVal)) {
        ListWrapper.forEach(rawClassVal,
            (className) => this._toggleClass(className, !isCleanup));
      } else {
        StringMapWrapper.forEach(rawClassVal, (expVal, className) {
          if (expVal) this._toggleClass(className, !isCleanup);
        });
      }
    }
  }

  void _toggleClass(String className, enabled) {
    this._renderer.setElementClass(this._ngEl, className, enabled);
  }
}
