library angular2.src.directives.ng_style;

import "package:angular2/metadata.dart" show Directive, LifecycleEvent;
import "package:angular2/core.dart" show ElementRef;
import "package:angular2/change_detection.dart"
    show KeyValueDiffer, KeyValueDiffers;
import "package:angular2/src/facade/lang.dart" show isPresent, isBlank, print;
import "package:angular2/src/render/api.dart" show Renderer;

/**
 * Adds or removes styles based on an {expression}.
 *
 * When the expression assigned to `ng-style` evaluates to an object, the corresponding element
 * styles are updated. Style names to update are taken from the object keys and values - from the
 * corresponding object values.
 *
 * # Example:
 *
 * ```
 * <div ng-style="{'text-align': alignEpr}"></div>
 * ```
 *
 * In the above example the `text-align` style will be updated based on the `alignEpr` value
 * changes.
 *
 * # Syntax
 *
 * - `<div ng-style="{'text-align': alignEpr}"></div>`
 * - `<div ng-style="styleExp"></div>`
 */
@Directive(
    selector: "[ng-style]",
    lifecycle: const [LifecycleEvent.onCheck],
    properties: const ["rawStyle: ng-style"])
class NgStyle {
  KeyValueDiffers _differs;
  ElementRef _ngEl;
  Renderer _renderer;
  var _rawStyle;
  KeyValueDiffer _differ;
  NgStyle(this._differs, this._ngEl, this._renderer) {}
  set rawStyle(v) {
    this._rawStyle = v;
    if (isBlank(this._differ) && isPresent(v)) {
      this._differ = this._differs.find(this._rawStyle).create(null);
    }
  }

  onCheck() {
    if (isPresent(this._differ)) {
      var changes = this._differ.diff(this._rawStyle);
      if (isPresent(changes)) {
        this._applyChanges(changes);
      }
    }
  }

  void _applyChanges(dynamic changes) {
    changes.forEachAddedItem((record) {
      this._setStyle(record.key, record.currentValue);
    });
    changes.forEachChangedItem((record) {
      this._setStyle(record.key, record.currentValue);
    });
    changes.forEachRemovedItem((record) {
      this._setStyle(record.key, null);
    });
  }

  void _setStyle(String name, String val) {
    this._renderer.setElementStyle(this._ngEl, name, val);
  }
}
