library angular2.src.router.instruction;

import "package:angular2/src/facade/collection.dart"
    show Map, MapWrapper, Map, StringMapWrapper, List, ListWrapper;
import "package:angular2/src/facade/lang.dart" show isPresent, isBlank, Type;
import "package:angular2/src/facade/async.dart" show Future;
import "path_recognizer.dart" show PathRecognizer;
import "url_parser.dart" show Url;

class RouteParams {
  Map<String, String> params;
  RouteParams(this.params) {}
  String get(String param) {
    return StringMapWrapper.get(this.params, param);
  }
}

/**
 * `Instruction` is a tree of `ComponentInstructions`, with all the information needed
 * to transition each component in the app to a given route, including all auxiliary routes.
 *
 * This is a public API.
 */
class Instruction {
  ComponentInstruction component;
  Instruction child;
  Map<String, Instruction> auxInstruction;
  Instruction(this.component, this.child, this.auxInstruction) {}
  Instruction replaceChild(Instruction child) {
    return new Instruction(this.component, child, this.auxInstruction);
  }
}

/**
 * Represents a partially completed instruction during recognition that only has the
 * primary (non-aux) route instructions matched.
 *
 * `PrimaryInstruction` is an internal class used by `RouteRecognizer` while it's
 * figuring out where to navigate.
 */
class PrimaryInstruction {
  ComponentInstruction component;
  PrimaryInstruction child;
  List<Url> auxUrls;
  PrimaryInstruction(this.component, this.child, this.auxUrls) {}
}

String stringifyInstruction(Instruction instruction) {
  var params = instruction.component.urlParams.length > 0
      ? ("?" + instruction.component.urlParams.join("&"))
      : "";
  return instruction.component.urlPath +
      stringifyAux(instruction) +
      stringifyPrimary(instruction.child) +
      params;
}

String stringifyPrimary(Instruction instruction) {
  if (isBlank(instruction)) {
    return "";
  }
  var params = instruction.component.urlParams.length > 0
      ? (";" + instruction.component.urlParams.join(";"))
      : "";
  return "/" +
      instruction.component.urlPath +
      params +
      stringifyAux(instruction) +
      stringifyPrimary(instruction.child);
}

String stringifyAux(Instruction instruction) {
  var routes = [];
  StringMapWrapper.forEach(instruction.auxInstruction, (auxInstruction, _) {
    routes.add(stringifyPrimary(auxInstruction));
  });
  if (routes.length > 0) {
    return "(" + routes.join("//") + ")";
  }
  return "";
}

/**
 * A `ComponentInstruction` represents the route state for a single component. An `Instruction` is
 * composed of a tree of these `ComponentInstruction`s.
 *
 * `ComponentInstructions` is a public API. Instances of `ComponentInstruction` are passed
 * to route lifecycle hooks, like [CanActivate].
 */
class ComponentInstruction {
  String urlPath;
  List<String> urlParams;
  PathRecognizer _recognizer;
  Map<String, dynamic> params;
  bool reuse = false;
  ComponentInstruction(this.urlPath, this.urlParams, this._recognizer,
      [this.params = null]) {}
  get componentType {
    return this._recognizer.handler.componentType;
  }

  Future<Type> resolveComponentType() {
    return this._recognizer.handler.resolveComponentType();
  }

  get specificity {
    return this._recognizer.specificity;
  }

  get terminal {
    return this._recognizer.terminal;
  }

  Object routeData() {
    return this._recognizer.handler.data;
  }
}
