library angular2.src.router.router_outlet;

import "package:angular2/src/facade/async.dart" show Future, PromiseWrapper;
import "package:angular2/src/facade/collection.dart" show StringMapWrapper;
import "package:angular2/src/facade/lang.dart" show isBlank, isPresent;
import "../core/metadata.dart" show Directive, Attribute;
import "package:angular2/core.dart"
    show DynamicComponentLoader, ComponentRef, ElementRef;
import "package:angular2/di.dart" show Injector, bind, Dependency, UNDEFINED;
import "router.dart" as routerMod;
import "instruction.dart" show Instruction, ComponentInstruction, RouteParams;
import "route_config_impl.dart" show ROUTE_DATA;
import "lifecycle_annotations.dart" as hookMod;
import "route_lifecycle_reflector.dart" show hasLifecycleHook;

/**
 * A router outlet is a placeholder that Angular dynamically fills based on the application's route.
 *
 * ## Use
 *
 * ```
 * <router-outlet></router-outlet>
 * ```
 */
@Directive(selector: "router-outlet")
class RouterOutlet {
  ElementRef _elementRef;
  DynamicComponentLoader _loader;
  routerMod.Router _parentRouter;
  routerMod.Router childRouter = null;
  String name = null;
  ComponentRef _componentRef = null;
  ComponentInstruction _currentInstruction = null;
  RouterOutlet(this._elementRef, this._loader, this._parentRouter,
      @Attribute("name") String nameAttr) {
    if (isPresent(nameAttr)) {
      this.name = nameAttr;
    }
    this._parentRouter.registerOutlet(this);
  }
  /**
   * Given an instruction, update the contents of this outlet.
   */
  Future<dynamic> commit(Instruction instruction) {
    instruction = this._getInstruction(instruction);
    var componentInstruction = instruction.component;
    if (isBlank(componentInstruction)) {
      return PromiseWrapper.resolve(true);
    }
    var next;
    if (componentInstruction.reuse) {
      next = this._reuse(componentInstruction);
    } else {
      next = this
          .deactivate(instruction)
          .then((_) => this._activate(componentInstruction));
    }
    return next.then((_) => this._commitChild(instruction));
  }

  Instruction _getInstruction(Instruction instruction) {
    if (isPresent(this.name)) {
      return instruction.auxInstruction[this.name];
    } else {
      return instruction;
    }
  }

  Future<dynamic> _commitChild(Instruction instruction) {
    if (isPresent(this.childRouter)) {
      return this.childRouter.commit(instruction.child);
    } else {
      return PromiseWrapper.resolve(true);
    }
  }

  Future<dynamic> _activate(ComponentInstruction instruction) {
    var previousInstruction = this._currentInstruction;
    this._currentInstruction = instruction;
    var componentType = instruction.componentType;
    this.childRouter = this._parentRouter.childRouter(componentType);
    var bindings = Injector.resolve([
      bind(ROUTE_DATA).toValue(instruction.routeData()),
      bind(RouteParams).toValue(new RouteParams(instruction.params)),
      bind(routerMod.Router).toValue(this.childRouter)
    ]);
    return this
        ._loader
        .loadNextToLocation(componentType, this._elementRef, bindings)
        .then((componentRef) {
      this._componentRef = componentRef;
      if (hasLifecycleHook(hookMod.onActivate, componentType)) {
        return this
            ._componentRef
            .instance
            .onActivate(instruction, previousInstruction);
      }
    });
  }

  /**
   * Called by Router during recognition phase
   */
  Future<bool> canDeactivate(Instruction nextInstruction) {
    if (isBlank(this._currentInstruction)) {
      return PromiseWrapper.resolve(true);
    }
    var outletInstruction = this._getInstruction(nextInstruction);
    if (hasLifecycleHook(
        hookMod.canDeactivate, this._currentInstruction.componentType)) {
      return PromiseWrapper.resolve(this._componentRef.instance.canDeactivate(
          isPresent(outletInstruction) ? outletInstruction.component : null,
          this._currentInstruction));
    }
    return PromiseWrapper.resolve(true);
  }

  /**
   * Called by Router during recognition phase
   */
  Future<bool> canReuse(Instruction nextInstruction) {
    var result;
    var outletInstruction = this._getInstruction(nextInstruction);
    var componentInstruction = outletInstruction.component;
    if (isBlank(this._currentInstruction) ||
        this._currentInstruction.componentType !=
            componentInstruction.componentType) {
      result = false;
    } else if (hasLifecycleHook(
        hookMod.canReuse, this._currentInstruction.componentType)) {
      result = this
          ._componentRef
          .instance
          .canReuse(componentInstruction, this._currentInstruction);
    } else {
      result = componentInstruction == this._currentInstruction ||
          (isPresent(componentInstruction.params) &&
              isPresent(this._currentInstruction.params) &&
              StringMapWrapper.equals(componentInstruction.params,
                  this._currentInstruction.params));
    }
    return PromiseWrapper.resolve(result).then((result) {
      // TODO: this is a hack
      componentInstruction.reuse = result;
      return result;
    });
  }

  Future<dynamic> _reuse(ComponentInstruction instruction) {
    var previousInstruction = this._currentInstruction;
    this._currentInstruction = instruction;
    return PromiseWrapper.resolve(hasLifecycleHook(
            hookMod.onReuse, this._currentInstruction.componentType)
        ? this._componentRef.instance.onReuse(instruction, previousInstruction)
        : true);
  }

  Future<dynamic> deactivate(Instruction nextInstruction) {
    var outletInstruction = this._getInstruction(nextInstruction);
    var componentInstruction =
        isPresent(outletInstruction) ? outletInstruction.component : null;
    return (isPresent(this.childRouter)
        ? this.childRouter.deactivate(
            isPresent(outletInstruction) ? outletInstruction.child : null)
        : PromiseWrapper.resolve(true)).then((_) {
      if (isPresent(this._componentRef) &&
          isPresent(this._currentInstruction) &&
          hasLifecycleHook(
              hookMod.onDeactivate, this._currentInstruction.componentType)) {
        return this
            ._componentRef
            .instance
            .onDeactivate(componentInstruction, this._currentInstruction);
      }
    }).then((_) {
      if (isPresent(this._componentRef)) {
        this._componentRef.dispose();
        this._componentRef = null;
      }
    });
  }
}
